/* Copyright (C) Bram Avontuur (bram@avontuur.org) */
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif
#include	<stdio.h>
#include	<unistd.h>
#ifdef		HAVE_STDLIB_H
#include	<stdlib.h>
#endif		/* HAVE_STDLIB_H */
#include	<errno.h>
#include	<ctype.h>
#include	<fcntl.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>

/* Custom headerfiles */

#include <getopt.h>
#include <avifile.h>
#include <videoencoder.h>
#include "capproc.h"
#include "v4lxif.h"
#include "global.h"
#include "parseconfig.h"
#include "channel.h"

/* External variables/functions */

/* Function prototypes */

void init_v4l();
void vcr_debug(const char*);
void usage();
void parse_arguments(int, char*argv[]);
void init_debug();
void init_globals();
void verify_globals();
void end_program(int);
short add_attribute(const char*);
short set_attributes();
int get_codec_index(int);
char **split_string(const char *);
short parse_config_file();
short set_error(const char *);

/* Global variables */

//enum cmdopts { OPT_
struct codec_attr {
	char *name;
	char *val;
} *attributes;

v4lxif *v4l;
FILE
	*vcr_debug_file = NULL;
short
	want_debug;
char
	*prog_name = NULL,
	*codec_name = NULL,
	*input_source,
	*preset,
	*filename,
	*config_file,
	*norm,
	*error = NULL;
int
	rectime,
	codecID,
	keyframes,
	quality,
	attr_count = 0;
vector<AttributeInfo>
	encoder_info;

int
main(int argc, char *argv[])
{
	prog_name = strdup(argv[0]); //let's hope it stays intact.

	init_globals();

	parse_arguments(argc, argv);

	if (want_debug)
		init_debug();

	if (parse_config_file() < 0)
	{
		fprintf(stderr, "Error parsing config file.: %s\n", error);
		exit(1);
	}

	//initialize video4linux interface[s]
	init_v4l();

	verify_globals();

	CaptureProcess *cap = new CaptureProcess(
		v4l,
		filename,
		-1,
		codecID,
		quality * 100,
		keyframes,
		(CaptureProcess::Sound_Freqs)0,
		(CaptureProcess::Sample_Sizes)0,
		(CaptureProcess::Sound_Chans)0,
		(CaptureProcess::Resolutions)0,
		rectime,
		-1,
		25.0);

	short ready = 0;
	while (!ready)
	{
		int
			processed,
			cap_drop,
			comp_drop,
			finished;
    char str[256];

		cap->getState(processed, cap_drop, comp_drop, finished);

    sprintf(str, "Captured: %d frames, dropped in capture: %d, dropped "\
			"in encoder: %d\n", processed, cap_drop, comp_drop);
		vcr_debug(str);

		ready = finished;
		sleep(1);
	}

	end_program(0);
}

void
end_program(int exit_status)
{
	v4l->setCapture(0);
	v4l->setAudioMute(1);
	delete v4l;

	exit(exit_status);
}

void
list_norms()
{
	fprintf(stderr, "Available norms: ");
	int i = 0;

	while (norms[i].nr > -1)
		fprintf(stderr, "%s ", norms[i].str);
	
	fprintf(stderr, "\n");
}

void
list_presets()
{
	fprintf(stderr, "Available presets are:\n\n");

	char **list;

	for (list = cfg_list_sections(); *list != NULL; list++)
	{
		if (!strcmp(*list, "defaults") ||
			!strcmp(*list, "global") ||
			!strcmp(*list, "launch"))
			continue;
		fprintf(stderr, "\"%s\" ", *list);
	}
	fprintf(stderr, "\n");
}

void
list_sources()
{
	int 
		i,
		nrsources;
	const char
		*srcname;

	fprintf(stderr, "Available input sources: ");
 
 	nrsources = v4l->capCapChannelC();
	for (i = 0; i < nrsources; i++)
	{
		fprintf(stderr, "%s", v4l->capChannelName(i));
		if (i != nrsources -1)
			fprintf(stderr, ", ");
		else
			fprintf(stderr, ".");
	}
	fprintf(stderr, "\n");
}

void
list_codecs()
{
	std::vector<CodecInfo>::iterator it;

	fprintf(stderr, "Available codecs: \n\n");

	for ( it = video_codecs.begin(); it != video_codecs.end(); it++)
	{
		if ( it->kind == CodecInfo::DShow_Dec)
			continue; //probably not a usable codec..
		
		const char *cname = it->text.c_str();
		int fourcc = it->fourcc;
		fprintf(stderr, "\"%s\"\n", cname);
	}
	fprintf(stderr, "\n");
}

void
list_attributes(int fourcc)
{
	int defval;
	char def_str[256];

	int idx = get_codec_index(fourcc);
	if (idx < 0)
		return;

	vector<AttributeInfo> encinfo = video_codecs[idx].encoder_info;

	fprintf(stderr, "These attributes are supported for this codec:\n\n");

	std::vector<AttributeInfo>::const_iterator it;
	for(it=encinfo.begin(); it!=encinfo.end(); it++)
	{
		vcr_debug("Looping through attributes...\n");
		fprintf(stderr, "Attribute \"%s\"\n", it->name.c_str());
		fflush(stderr);
		switch(it->kind)
		{
		case AttributeInfo::Integer:
			IVideoEncoder::GetExtendedAttr(fourcc, it->name.c_str(), defval);
			fprintf(stderr, "\tType: integer (default value: %d)\n", defval);
			break;
		case AttributeInfo::Select:
			IVideoEncoder::GetExtendedAttr(fourcc, it->name.c_str(), defval);
			fprintf(stderr, "\tType: select (default value: %s)\n",
				it->options[defval].c_str());
			fprintf(stderr, "\tPossible values: ");

			std::vector<std::string>::const_iterator sit;
			for (sit=(it->options).begin(); sit!=(it->options).end(); sit++)
				fprintf(stderr, "\"%s\" ", sit->c_str());

			fprintf(stderr, "\n");
			break;
		case AttributeInfo::String:
			IVideoEncoder::GetExtendedAttr(fourcc, it->name.c_str(), def_str, 256);
			fprintf(stderr, "\tType: string (default value: %s)\n", def_str);
			break;
		}
		fprintf(stderr, "\n");
	}
}

void
init_v4l()
{
	v4l = new v4l1if(0, "/dev/video");
	v4l->setAudioMute(0);
	v4l->setChannel(1);
	//list_sources();
}

void
vcr_debug(const char *txt)
{
	if (vcr_debug_file)
	{
		fwrite(txt, sizeof(char), strlen(txt), vcr_debug_file);
		fflush(vcr_debug_file);
	}
}

void
usage()
{
	fprintf(stderr, 
		"Usage: %s [options] filename\n\n"\
		"Options can be one or more of the following:\n\n"\
		"--codec-attribute, -a <attribute>=<value>\n"\
		"\tSets a codec attribute. To get a list of attributes for your \n"\
		"\tselected codec, give 'list=1' as argument. You can set more than\n"\
		"\tone attribute by using this option multiple times.\n\n"\
		"--codec, -c <codec name>\n"\
		"\tWhich codec to use (specify a bad string to get a list of\n" \
		"\tavailable codecs)\n\n"\
		"--debug, -d\n"\
		"\tTurn on debug info (which will be logged in ~/.vcrdebug)\n\n"\
		"--config-file, -f\n"\
		"\tWhich file to use as configuration file (default= ~/.vcrrc)\n"\
		"\tRead 'man vcrrc' for the syntax\n\n"\
		"--preset, -p <tuner preset>\n"\
		"\tWhich tuner preset to use. Tuner presets are read from the config\n"\
		"\tfile, and have the same syntax/setup as xawtv's presets (so read \n"\
		"\tthe xawtv manual how to make/generate one!)\n\n"\
		"--keyframes, -k <keyframes>\n"\
		"\tNumber of frames between keyframes (default=15)\n\n"\
		"--quality, -q <quality>\n"\
		"\tQuality for codec (1..100, 100 being best, defaults = 95)\n\n"\
		"--rectime, -r <time>\n"\
		"\tLength of time this program will record (specify in seconds)\n\n"\
		"--source, -s <source>\n"\
		"\tInput source (specify a bad string to get a list of available "\
		"sources)\n\n"\
		"", prog_name);
}

void
init_globals()
{
	input_source = NULL;
	want_debug = 0;
#if 0
	rectime = 10; //default 10 sec. recording time.
	codec_name = (char*)malloc(80 * sizeof(char));
	strcpy(codec_name, "Indeo Video 5.0");
	codecID = fccIV50;
	keyframes = 15;
	quality = 95;
#else
	rectime = 0;
	codec_name = NULL;
	codecID = -1;
	keyframes = 0;
	quality = 0;
#endif
	preset = NULL;
	filename = NULL;
	attributes = NULL;
	attr_count = 0;
	config_file = NULL;
	error = NULL;
	norm = NULL;
}

void
init_debug()
{
	char *homedir = get_homedir(NULL);
	char *dbfile = new char[strlen(homedir) + 255];

	strcpy(dbfile, homedir);
	free(homedir);
	strcat(dbfile, "/.vcrdebug");
	vcr_debug_file = fopen(dbfile, "a");
	delete[] dbfile;

	if (!vcr_debug_file)
	{
		fprintf(stderr, "Could not open debug file.\n");
		exit(1);
	}
	vcr_debug("VCR Debugging started!\n");
}

int
is_valid_norm(const char *n)
{
	if (!n)
		return -1;

	int i = 0;

	while (norms[i].nr > -1)
	{
		if (!strcasecmp(norms[i].str, n))
		{
			vcr_debug("Found valid channel norm: ");
			vcr_debug(norms[i].str);
			vcr_debug("\n");
			return norms[i].nr;
		}
		i++;
	}
	return -1;
}

/* returns -1 if codec does not exist.
 * Otherwise, the codec's unique ID
 */
int
is_valid_codec(const char *cname)
{
	if (!cname)
		return -1;

	int
		found_codec = -1;

	std::vector<CodecInfo>::iterator it;

	for (it = video_codecs.begin(); it != video_codecs.end(); it++)
	{
		if ( it->kind == CodecInfo::DShow_Dec)
			continue;

		if (!strcasecmp(cname, it->text.c_str()))
		{
			found_codec = it->fourcc;
			encoder_info = it->encoder_info;
			vcr_debug("Codec gevonden!\n");
			vcr_debug(cname);
			vcr_debug("\n");
			break;
		}
	}

	return found_codec;
}

/* Gets the 'channel' and 'fine' settings for the preset from the xawtv,
 * calculates the frequency using those 2 values, and sets the tuner.
 */
short
set_tuner_frequency(const char *preset)
{
	
	vcr_debug("Preset to find: ");
	vcr_debug(preset);
	vcr_debug("\n");
	char
		*cname = cfg_get_str(preset, "channel");
	int
		fine = cfg_get_int(preset, "fine");
	
	if (!cname || !strcmp(cname, "none"))
		return 0;

	int
		channel = lookup_channel(cname);

	if (channel < 0)
		return 0;

	int freq = get_freq(channel);
	
	if (freq < 0)
		return 0;
	
	freq += fine;

	v4l->setFreq(freq);

	vcr_debug("Set tuner frequency successfully.\n");
	return 1;
}

int
get_codec_index(int fourcc)
{
	std::vector<CodecInfo>::iterator it;
	int i = 0;

	for (it = video_codecs.begin(); it != video_codecs.end(); it++)
	{
		if (it->fourcc == fourcc)
			return i;
		i++;
	}

	return -1;
}

//TODO: It ignores fourcc currently..
short
set_attribute(int fourcc, const char *attr, const char *val)
{
	int setval = 0;
	int retval = 0;

	char fiepje[255];
	sprintf(fiepje, "set_attribute(%s, %s)\n", attr, val);
	vcr_debug(fiepje);

	std::vector<AttributeInfo>::const_iterator it;
	int idx = get_codec_index(fourcc);
	if (idx < 0)
		return 0;

	vector<AttributeInfo> encinfo = video_codecs[idx].encoder_info;

	for (it = encinfo.begin(); it != encinfo.end(); it++)
	{
		if (strcasecmp(attr, it->name.c_str()))
			continue; //wrong attribute.

		vcr_debug("Geldig attribuut gevonden!\n");
		switch(it->kind)
		{
		case AttributeInfo::Integer:
			vcr_debug("Het is een integer attribuut\n");
			IVideoEncoder::SetExtendedAttr(fourcc, it->name.c_str(), atoi(val));
			break;
		case AttributeInfo::String:
			IVideoEncoder::SetExtendedAttr(fourcc, it->name.c_str(), val);
			break;
		case AttributeInfo::Select:
			IVideoEncoder::SetExtendedAttr(fourcc, it->name.c_str(), val);
			break;
		}
		retval = 1;
		break;
	}
	return retval;
}

short
set_attributes(int fourcc)
{
	const char *cname = NULL;
	int i;

	char fiepje[255];
	sprintf(fiepje, "set_attributes(%d), attr_count: %d\n", fourcc, attr_count);
	vcr_debug(fiepje);

	for (i = 0; i < attr_count; i++)
	{
		if (!set_attribute(fourcc, attributes[i].name, attributes[i].val))
			return 0;
	}

	return 1;
}

int
is_valid_preset(const char *p)
{
	if (!p)
		return 0;

	int
		found_preset = 0;
	char **list;

	for (list = cfg_list_sections(); *list != NULL; list++)
	{
		if (!strcmp(*list, "defaults") ||
			!strcmp(*list, "global") ||
			!strcmp(*list, "launch") ||
			strcasecmp(*list, preset))
			continue;

		found_preset = 1;
		break;
	}

	vcr_debug("Found valid preset.\n");
	return found_preset;
}

/* returns -1 if inputsource does not exist.
 * Otherwise, the inputsource's ID (for v4l->setChannel())
 */
int
is_valid_inputsource(const char *src)
{
	if (!src)
		return -1;
	
	int
		i,
	 	nrsources = v4l->capCapChannelC();
	int
		found_channel = -1;

	for (i = 0; i < nrsources; i++)
	{
		if (!strcasecmp(v4l->capChannelName(i), src))
		{
			found_channel = i;
		}
	}

	return found_channel;
}

/* Side effect: mangles argv */
void
parse_arguments(int argc, char *argv[])
{
	int
		c,
		long_index;
	static struct option long_options[] = 
		{
			/*
			{ "downsample", 0, 0, '2'},
			{ "8bits", 0, 0, '8'},
			{ "autolist", 1, 0, 'a'},
			{ "config-file", 1, 0, 'c'},
			{ "debug", 0, 0, 'd'},
			{ "help", 0, 0, 'h'},
			{ "list", 1, 0, 'l'},
			{ "no-mixer", 0, 0, 'n'},
			{ "chroot", 1, 0, 'o'},
			{ "playmode", 1, 0, 'p'},
			{ "dont-quit", 0, 0, 'q'},
			{ "runframes", 1, 0, 'r'},
			{ "sound-device", 1, 0, 's'},
			{ "threads", 1, 0, 't'},
			{ 0, 0, 0, 0}
			*/
			{ "codec-attribute", 1, 0, 'a'},
			{ "codec", 1, 0, 'c'},
			{ "debug", 0, 0, 'd'},
			{ "config-file", 1, 0, 'f'},
			{ "keyframes", 1, 0, 'k'},
			{ "preset", 1, 0, 'p'},
			{ "quality", 1, 0, 'q'},
			{ "rectime", 1, 0, 'r'},
			{ "source", 1, 0, 's' },
			{ 0, 0, 0, 0 }
		};
		
	while(1)
	{
		c = getopt_long_only(argc, argv, "a:c:df:k:p:q:r:s:", long_options,
			&long_index);

		if (c == EOF)
			break;

		switch(c)
		{
		case ':':
		case '?':
			usage();
			exit(1);
			break;
		case 'a': /* codec attributes */
			if (!add_attribute(optarg))
			{
				usage();
				exit(1);
			}
			break;
		case 'c': /* codec */
			if (codec_name)
				free(codec_name);
			codec_name = strdup(optarg);
			break;
		case 'd': /* debug */
			want_debug = 1;
			break;
		case 'f': /* config-file */
			if (config_file)
				free(config_file);
			config_file = strdup(optarg);
			break;
		case 'k': /* keyframes */
			keyframes = atoi(optarg);
			break;
		case 'p': /* tuner preset (taken from xawtv config file) */
			if (preset)
				free(preset);
			preset = strdup(optarg);
			break;
		case 'q': /* quality */
			quality = atoi(optarg);
			break;
		case 'r': /* recording time */
			rectime = atoi(optarg);
			break;
		case 's': /* input source */
			if (input_source)
				free(input_source);
			input_source = strdup(optarg);
			break;
		default:
			usage();
			exit(1);
		}
	}

	if (optind < argc) /* remaining argument = filename */
	{
		if (argc - optind != 1)
		{
			usage();
			exit(1);
		}
		if (filename)
			free(filename);
		filename = strdup(argv[optind]);
	}
}

/* called after reading config file & parsing cmdline args */
void
verify_globals()
{
	int temp = 0;

	if (!filename)
	{
		fprintf(stderr, "No filename given!\n");
		usage();
		end_program(1);
	}

	if (!quality)
		quality = 95;
	if (quality < 1 || quality > 100)
	{
		fprintf(stderr, "Quality out of range (range is 1..100, default 95)\n");
		end_program(1);
	}

	if (!keyframes)
		keyframes = 15;
	if (keyframes < 1 || keyframes > 30)
	{
		fprintf(stderr, "Keyframes out of range (range is 1..30, default 15)\n");
		end_program(1);
	}

	if (norm && (temp = is_valid_norm(norm)) < 0)
	{
		fprintf(stderr, "Invalid norm given!\n");
		list_norms();
		end_program(1);
	}
	else if (norm)
		v4l->setChannelNorm(temp);

	if ((temp = is_valid_inputsource(input_source)) < 0)
	{
			fprintf(stderr, "Invalid input source given!\n");
			list_sources();
			end_program(1);
	}
	else
		v4l->setChannel(temp);

	if (!codec_name)
	{
		codec_name = (char*)malloc(80 * sizeof(char));
		strcpy(codec_name, "Indeo Video 5.0");
	}

	if ((temp = is_valid_codec(codec_name)) < 0)
	{
		fprintf(stderr, "Invalid codec name given!\n");
		list_codecs();
		end_program(1);
	}
	else
		codecID = temp;
	
	if (attr_count && !set_attributes(codecID))
	{
		fprintf(stderr, "Invalid attribute[s] set for this codec!\n");
		list_attributes(codecID);
		end_program(1);
	}

	if (preset && !(temp = is_valid_preset(preset)))
	{
		fprintf(stderr, "Invalid tuner preset given! (Couldn't find it in "\
			"your config file)\n");
		list_presets();
		end_program(1);
	}
	else if (preset && !set_tuner_frequency(preset))
	{
		fprintf(stderr, "Couldn't set tuner for chosen preset (maybe no channel"\
			" is set in the .xawtv config file?)\n");
		end_program(1);
	}

	if (!rectime)
	{
		fprintf(stderr, "Recording time should be -1 for unlimited time, else "\
			"bigger than zero.\n");
		end_program(1);
	}
}

short
add_attribute(const char *attr)
{
	int 
		attr_len = strlen(attr) + 1;
	char
		attrname[attr_len],
		attrval[attr_len];
	struct codec_attr
		new_attr;

	int val = sscanf(attr,"%[^=]=%s", attrname, attrval);
	if (val != 2)
		return 0;
	
	new_attr.name = strdup(attrname);
	new_attr.val = strdup(attrval);

	attributes = (struct codec_attr*)realloc(attributes, (attr_count + 1) *
		sizeof(struct codec_attr));
	attributes[attr_count++] = new_attr;	

	return 1;
}

short
parse_config_file()
{
	const char *tmpstr;
	int tmpint;
	char *tmperr;

	if (!config_file)
	{
		char * homedir = get_homedir(NULL);
		if (homedir)
		{
			config_file = (char*)malloc((strlen(homedir) + 20) * sizeof(char));
			sprintf(config_file, "%s/.vcrrc", homedir);
			free(homedir);
		}
		else
			return set_error("Could not find your homedir to search for default config "\
				"file.");
	}
	if (cfg_parse_file(config_file) < 0)
		return -1;

	if (!attr_count && (tmpstr = cfg_get_str("defaults", "attributes")))
	{
		vcr_debug("Attributes: ");
		vcr_debug(tmpstr);
		vcr_debug("\n");

		char **attrs = split_string(tmpstr);
		if (!attrs)
			return set_error("Couldn't parse 'attributes' in config file.");

		int i = 0, success = 1;
		while (attrs[i])
		{
			if (!add_attribute(attrs[i]))
			{
				success = 0;
				tmperr = new char[strlen(attrs[i])+100];
				sprintf(tmperr, "Couldn't add attribute '%s' from config file.", attrs[i]);
				set_error(tmperr);
				free(attrs[i]);
				break;
			}
			free(attrs[i]);
			i++;
		}
		free(attrs);
		if (!success)
			return -1;
	}

	if (!codec_name && (tmpstr = cfg_get_str("defaults", "codec")))
		codec_name = strdup(tmpstr);
	
	if (!keyframes && (tmpint = cfg_get_int("defaults", "keyframes")) > -1)
		keyframes = tmpint;
	
	if (!norm && (tmpstr = cfg_get_str("defaults", "norm")))
		norm = strdup(tmpstr);

	if (!quality && (tmpint = cfg_get_int("defaults", "quality")) > -1)
		quality = tmpint;
	
	if (!input_source && (tmpstr = cfg_get_str("defaults", "source")))
		input_source = strdup(tmpstr);
	

	return 0;
}

char **
split_string(const char *tmpvals)
{
	unsigned int
		i = 0, curval_pos = 0, val_index = 0;
	char
		**values = NULL,
		*tmpval2 = NULL,
		prev = '\0';
	
	
	if (!tmpvals)
	{
		values = (char**)realloc(values, sizeof(char*));
		values[0] = NULL; //last element in the array.
		return values;
	}

	values = (char**)malloc(1 * sizeof(char*));
	values[val_index] = (char*)calloc(strlen(tmpvals)+1, sizeof(char));
	while (i < strlen(tmpvals))
	{
		char curr = tmpvals[i];
		if (prev == '\\') 
		{
			//previous read char was a backslash. That implies that this 
			//character is treated a 'special' character. Currently only
			//comma and backslash itself.
			switch(curr)
			{
			case '\\':
			case ',':
				values[val_index][curval_pos++] = curr;
				break;
			}
			prev = '\0'; //prevent special char interpr. in next loop ;)
		}
		else if (curr == '\\') 
		{
			prev = '\\';
		}
		else if (curr == ',') //keyword separator
		{
			tmpval2 = crop_whitespace(values[val_index]);
			free(values[val_index]);
			values[val_index] = tmpval2;
			curval_pos = 0;
			values = (char**)realloc(values, ((++val_index)+1) * sizeof(char*));
			values[val_index] = (char*)calloc(strlen(tmpvals)+1, sizeof(char));
			prev = ',';
		}
		else //normal character
		{
			values[val_index][curval_pos++] = curr;
			prev = curr;
		}
		i++;
	}

	tmpval2 = crop_whitespace(values[val_index]);
	free(values[val_index]);
	values[val_index] = tmpval2;
	values = (char**)realloc(values, ((++val_index)+1) * sizeof(char*));
	values[val_index] = NULL; //last element in the array.

	return values;
}

/* sets global error var and returns -1 */
short
set_error(const char*txt)
{
	if (!txt)
		return -1;

	error = new char[strlen(txt)+1];
	strcpy(error, txt);
	return -1;
}
